require 'tk'

#
# This application handles special formatting cautions that are currently in place for certain Montana forms
# for the 2014 season. Refer to form CDCH1007 for an example. The output from this application is to be
# inserted into the form code at the very end of the file. This way, the calcs above this special code will
# function as normal but the SuperForm view will be based on the formatting code produced by this application.
#

#
# #########################################################
#
# Message Box section:
#

require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

# Window drawing section.

root = TkRoot.new() { title "DISPLAY FORMATTER PROGRAM" }
root.geometry('+225+250')

file_name_label = TkLabel.new(root, :text=>"", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:side=>"top", :fill=>:x, :padx=>10, :pady=>10)

#
# Main frames.
#
setup_frame = TkFrame.new(root, :borderwidth=>5, :relief=>:ridge).pack(:side=>"top", :fill=>"x", :padx=>10)

control_button_frame = TkFrame.new(root, :borderwidth=>5, :relief=>:ridge).pack(:side=>"top", :fill=>"x", :padx=>10)

output_frame = TkFrame.new(root, :borderwidth=>5, :relief=>:ridge).pack(:side=>"top", :fill=>"both", :expand=>true, :padx=>10, :pady=>10)

#
# list frames.
#
list_frame2 = TkFrame.new(setup_frame).pack("side"=>"left")
list_frame3 = TkFrame.new(setup_frame).pack("side"=>"left")
list_frame4 = TkFrame.new(setup_frame).pack("side"=>"left")

box_list_frame = TkFrame.new(list_frame2, :borderwidth=>3, :relief=>:ridge).pack("side"=>"top", "padx" =>10, "pady" =>10)
box_list_label = TkLabel.new(box_list_frame, :text=>"Calc Box List", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

box_delete_frame = TkFrame.new(box_list_frame).pack("side"=>"bottom", "pady" => 5)
box_delete_button = TkButton.new(box_delete_frame, :text=>"Delete", :background=>"lightgrey", :foreground=>"blue", :width=>10, "font"=>"arial 10").pack()

box_list_bar_vertical = TkScrollbar.new(box_list_frame).pack("side"=>"right", "fill"=>"y")
box_list = TkListbox.new(box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>10, "height"=>10, :selectmode=>"extended", "font"=>"arial 10").pack("side"=>"left")
box_list.yscrollbar(box_list_bar_vertical)

action_list_frame = TkFrame.new(list_frame4, :borderwidth=>3, :relief=>:ridge).pack("side"=>"top", "padx" =>10, "pady" =>10)
action_list_label = TkLabel.new(action_list_frame, :text=>"Display Formats", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

action_list_bar_vertical = TkScrollbar.new(action_list_frame).pack("side"=>"right", "fill"=>"y")
action_list = TkListbox.new(action_list_frame, :width=>55, "background"=>"white", "exportselection" => 0, "height"=>5, :selectmode=>"browse", "font"=>"arial 10").pack("side"=>"left")
action_list.yscrollbar(action_list_bar_vertical)

action_list.insert("end", "Enter 99.99 as 99 99 with a space between dollars and cents")
action_list.insert("end", "Enter 100.00 as 100 00 with a space between dollars and cents")
action_list.insert("end", "Enter 100.00 as 10000 with no space between dollars and cents")
action_list.insert("end", "Blank all calculated fields that are equal to zero")

#
# Default selection at startup.
#
action_list.selection_set(0)

generate_button = TkButton.new(control_button_frame, :text=>"Generate Code", :background=>"lightgrey", :foreground=>"blue", :width=>18, "font"=>"arial 10").pack(:side=>"left", :padx=>10, :pady=>5)

output_label = TkLabel.new(output_frame, :text=>"Generated Display Formatter Code - NOTE: ' Ctrl / ' selects all text.", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)
out_text_bar_vert = TkScrollbar.new(output_frame).pack("side"=>"right", "fill"=>"y")
out_text = TkText.new(output_frame, "height"=>15, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)
out_text.yscrollbar(out_text_bar_vert)

#
# Action Section:
#

#
# ###############################################################
#
# Globals:
#
cur_dir = "C:/Users/"

#
# This array holds the form boxes that are assigned a value, i.e. calculated.
#
calc_box_array = Array.new

def process_errors(type, error, var)

	err_msg = [
		"  ERROR! The number of calculated boxes does not match the number of printed boxes.\n",
		"  ERROR! Print statement boxes do not match  -  #{var}.\n",
		"  This form has no calculations to format.\t\n"
	]

	#
	# Using bitwise operations, error is set on a bit by bit basis, where each bit represents an error message.
	#
	if (error > 0)

		mask = 1

		#
		# Math.frexp(error) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of error. In our case, the fraction is 0.5 and the exponent is a power of 2, as our error is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say error = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new error.
		#
		0.upto(Math.frexp(error)[1]-1) do |shift|
			if (((mask << shift) & error) > 0)
				messageBox(err_msg[shift], type, $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

#
# This reads in the Ruby Code starting after $\="\x03\n" and looks for box assignments paired with the printing of those boxes.
# This will need to be updated with better calc assignment logic, for now this will do.
#
# format is the action_list selection - if we are blanking on zero, we want all calculated boxes.
#
def read_ruby_code(file, format)

	box_calc_array = Array.new
	box_print_array = Array.new
	box_non_calc_array = Array.new

	box = ''
	process = false

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			case
				when (line.delete("\s\t") =~ /\$\\="\\x03\\n"/): process = true

				when (process)

					case (line.delete("\s\t"))

						when (/print'(box\d+_\d+)',(box\d+_\d+)/)

							if ($1.eql?($2))

								unless ((box_print_array.include?($1)) or (box_non_calc_array.include?($1)))
									box_print_array.push($1)
								end
							else
								#
								# Probably an error here...
								#
								process_errors("Error!", 2, "print'#{$1}', #{$2}")
								process = false
							end

						when (/(box\d+_\d+)=/)

							box = $1
							rvalue = $'

							case (rvalue)

								when (/'%0\.[1,3-9]f'%/)

									unless (format == 3)

										#
										# We found a non-currency format string, remove the box from the box_calc_array
										# and from the box_print_array if it is already in there and insert the box into the
										# box_non_calc_array if it is not in there already.
										#
										box_calc_array.delete(box)
										box_print_array.delete(box)
										box_non_calc_array.push(box) unless (box_non_calc_array.include?(box))
									end

								when (/^(box\d+_\d+)(\.delete\(["']\*["']\))?$/)

									#
									# We found a post to a box from a box. We also look if it is a post from a barcode text box where we
									# delete the asterisk [$2]. We will only include the box we are posting TO in the box_calc_array only
									# if the box we are posting FROM is already in the box_calc_array.
									#
									if ($2)
										#
										# Posting from barcode text box with an asterisk.
										#
										box_non_calc_array.push(box) unless (box_non_calc_array.include?(box))
									else
										if (box_calc_array.include?($1))
											box_calc_array.push(box) unless (box_calc_array.include?(box))
										else
											#
											# A post from a non-calculated box like a name/address box.
											#
											box_non_calc_array.push(box) unless (box_non_calc_array.include?(box))
										end
									end
								else
									unless ((box_calc_array.include?(box)) or (box_non_calc_array.include?(box)))
										box_calc_array.push(box)
									end
							end
					end
			end
		end
	end

	if (box_calc_array.length != box_print_array.length)

		if (process)

			#
			# Error! They should be the same size.
			#
			process_errors("Error!", 1, nil)
		end

		return nil
	end

	return box_calc_array
end

def get_file_name(dir, type)

	ftypes = case (type)
			when ("RBW"): [ ["Ruby source files", '.rbw'] ]
		end

	Tk.getOpenFile(:filetypes=>ftypes, :initialdir=>dir)
end

generate_button.bind("Button-1") {

	out_text.delete("1.0", "end")

	#
	# Check first for an empty box list:
	#
	if (calc_box_array)

		if (calc_box_array.length > 0)

			#
			# First create the comment block to explain what is being done:
			#
			t = Time.now
			date = t.strftime("%m/%d/%y")
			action_str = action_list.get(action_list.index("active"))

			#
			# First we set the comment based on the selection in the action_list:
			#
			if ((0..2) === action_list.index("active"))

				str = "#\n# #{date}: Special $ field display code to meet jurisdiction popup caution: #{action_str}.\n# This special display code needs to be added at the end of the 'normal' calcs. The above calcs are calculated normally and we are displaying\n# those calculated values here to meet the caution. Note that the above calculations we are using here must still be formatted as per the\n# caution, that is, if the caution wants us to round a calculated result we must still round up in the main body of the calcs, we would simply\n# display those rounded values differently here.\n#\n"
			else

				str = "#\n# #{date}: Special $ field display code to blank all calculated fields that are equal to zero.\n# This special display code needs to be added at the end of the 'normal' calcs.\n#\n"
			end

			out_text.insert("end", str)

			#
			# Generating the format box list - we check the selection in the action_list to determine the name of the list:
			#
			list_type = ((0..2) === action_list.index("active") ? "format" : "blank")

			str = "#{list_type}_list = %w{#{calc_box_array.join(' ')}}\n\n"
			out_text.insert("end", str)

			#
			# Generating the formatting code for the boxes in the format_list based on Display Format selection.
			#
			case action_list.index("active")

				when (0): str = "format_list.each do |x|\n\tcode1 = <<ENDOFSTRING1\n\tprint x, #\{x}.sub(' ','.') unless (#\{x}.empty?)\nENDOFSTRING1\n\teval(code1)\nend"

				when (1): str = "format_list.each do |x|\n\tcode1 = <<ENDOFSTRING1\n\tprint x, #\{x}+' 00' unless (#\{x}.empty?)\nENDOFSTRING1\n\teval(code1)\nend"

				when (2): str = "format_list.each do |x|\n\tcode1 = <<ENDOFSTRING1\n\tprint x, #\{x}+'00' unless (#\{x}.empty?)\nENDOFSTRING1\n\teval(code1)\nend"

				when (3): str = "#{list_type}_list.each do |x| print x, eval(\"#\{x}.empty? ? #\{x} : #\{x}.delete(' ').to_f == 0.0 ? '' : #\{x}\") end"
			end

			out_text.insert("end", str)

			# Set the focus to the output text window, and put the cursor at the first char on line one.
			out_text.focus
			out_text.mark_set('insert', '1.0')
		end
	end
}

box_delete_button.bind("Button-1") {

	#
	# Delete the selected boxes from the list.
	#
	if (calc_box_array)

		box_list.curselection.each do |x| calc_box_array.delete(box_list.get(x)) end
		box_list.delete(0, "end")
		calc_box_array.each do |box| box_list.insert("end", box) end
	end
}

#
# Adding menu to this program. a work in process...................
#
#
# ###########################################################
#
# Menu:
#
#		Read:
#			Read Ruby Code file to create form box list [displaying box numbers].
#
#		Exit:
#			Exit the Application.
#
main_menu = TkMenu.new(root)

read_menu = TkMenu.new(main_menu, :tearoff=>false)

#
#
# This reads the .RBW file.
#
read_menu.add('command', :label=> "Ruby Source Code File", :command=> proc {

	tmp_array = Array.new

	fname = get_file_name(cur_dir, "RBW")

	unless (fname.empty?)

		#
		# Set File Label text and clear the out_text area:
		#
		file_name_label.configure("text", fname)
		out_text.delete("1.0", "end")

		cur_dir = File.dirname(fname)
		box_list.delete(0, "end")

		#
		# Now read and process the calculations if there are any:
		#
		calc_box_array = read_ruby_code(fname, action_list.index("active"))

		if (calc_box_array)

			#
			# Sort the array first - easier to look for non $ boxes:
			#
			calc_box_array.join.scan(/(box)(\d+)(_)(\d+)/) do tmp_array.push([$2.to_i,$4.to_i]) end
			tmp_array.sort!

			calc_box_array.clear
			tmp_array.each_index do |index| calc_box_array.push("box#{tmp_array[index][0]}_#{tmp_array[index][1]}") end
			calc_box_array.each { |box| box_list.insert("end", box) }
			process_errors("Attention!", 4, nil) unless (calc_box_array.length > 0)
		end
	end
})

main_menu.add('cascade', :menu=>read_menu, :label=>"Read")

profile_menu = TkMenu.new(main_menu, :tearoff=>false)

main_menu.add('command', :label=> "Exit", :command=> proc{exit})

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

Tk.mainloop()
